@@ -1,8 +1,10 @@
 """Production-grade runner with streaming, monitoring, and error handling."""
 
+import sys
 import time
 import signal
 import uuid
+import threading
 from concurrent.futures import ThreadPoolExecutor, as_completed
 from typing import Optional
 from tqdm import tqdm
@@ -103,40 +105,87 @@ class Runner:
         output_file = storage_config.get('output_file', f'workspaces/{self.workspace_id}/output.jsonl')
         partial_file = storage_config.get('partial_file', f'workspaces/{self.workspace_id}/partial.jsonl')
         failed_file = storage_config.get('failed_file', f'workspaces/{self.workspace_id}/failed.jsonl')
+        rejected_file = storage_config.get('rejected_file', f'workspaces/{self.workspace_id}/rejected.jsonl')
         
         self.incremental_saver = IncrementalSaver(
             output_file=output_file,
             partial_file=partial_file,
             failed_file=failed_file,
+            rejected_file=rejected_file,
             use_file_locking=True
         )
         
         # Initialize streaming data loader
         self.data_loader = StreamingConversationLoader(config, self.checkpoint_manager)
         
+        # Track for graceful shutdown - use Event for thread-safe shutdown
+        self.shutdown_event = threading.Event()
+        
         # Initialize generator with production components
         self.generator = ConversationGenerator(
             config=config,
             rate_limiter=self.rate_limiter,
             error_handler=self.error_handler,
-            incremental_saver=self.incremental_saver
+            incremental_saver=self.incremental_saver,
+            shutdown_event=self.shutdown_event
         )
         
-        # Track for graceful shutdown
-        self.shutdown_requested = False
         self._setup_signal_handlers()
         
         logger.info(f"Production runner initialized for workspace: {self.workspace_id}")
     
     def _setup_signal_handlers(self):
-        """Setup handlers for graceful shutdown."""
+        """Setup handlers for emergency shutdown."""
         def signal_handler(signum, frame):
-            logger.warning("\n⚠️  Shutdown signal received. Saving checkpoint...")
-            self.shutdown_requested = True
+            if not self.shutdown_event.is_set():
+                logger.warning("\n🛑 EMERGENCY SHUTDOWN INITIATED - Canceling all pending tasks...")
+                self.shutdown_event.set()
+            else:
+                # Second Ctrl+C = force immediate exit
+                logger.error("\n⚠️  FORCE EXIT - Shutting down immediately without saving!")
+                sys.exit(1)
         
         signal.signal(signal.SIGINT, signal_handler)
         signal.signal(signal.SIGTERM, signal_handler)
     
+    def _get_rate_limit_metric(self) -> str:
+        """
+        Get appropriate rate limit metric display based on limiter type.
+        
+        Returns:
+            Formatted string showing either:
+            - "Active: X/Y" for concurrency limiters
+            - "RPM: X" for traditional rate limiters
+        """
+        try:
+            all_stats = self.rate_limiter.get_all_stats()
+            
+            if not all_stats:
+                return "RPM:0"
+            
+            # Check first limiter to determine type
+            first_limiter_stats = next(iter(all_stats.values()))
+            
+            # ConcurrencyLimiter has 'active_calls' and 'max_concurrent'
+            if 'active_calls' in first_limiter_stats and 'max_concurrent' in first_limiter_stats:
+                # Aggregate concurrent calls across all limiters
+                total_active = sum(stats.get('active_calls', 0) for stats in all_stats.values())
+                total_max = sum(stats.get('max_concurrent', 0) for stats in all_stats.values())
+                return f"Active:{total_active}/{total_max}"
+            
+            # RateLimiter has 'current_rpm'
+            elif 'current_rpm' in first_limiter_stats:
+                # Aggregate RPM across all limiters
+                total_rpm = sum(stats.get('current_rpm', 0) for stats in all_stats.values())
+                return f"RPM:{total_rpm}"
+            
+            else:
+                return "RPM:0"
+                
+        except Exception as e:
+            logger.debug(f"Error getting rate limit metric: {e}")
+            return "RPM:0"
+    
     def run(self):
         """Run the pipeline with full production features."""
         try:
@@ -272,12 +321,65 @@ class Runner:
             except Exception as e:
                 logger.error(f"Error finalizing storage: {e}")
     
+    def _async_save_result(self, result: dict, status: str, position: int, content_hash: str, conv_id: int):
+        """
+        Async I/O: Save result to disk without blocking LLM workers.
+        
+        This runs in a separate I/O executor thread, ensuring LLM API calls
+        never wait for disk writes.
+        """
+        try:
+            # Save to incremental saver
+            self.incremental_saver.save_conversation(result, status=status)
+            
+            # Record to monitoring
+            if self.monitor:
+                try:
+                    processing_time = result.get('processing_time_ms', 0)
+                    tokens_total = result.get('tokens', {}).get('total_tokens', 0)
+                    self.monitor.record_conversation(
+                        conversation_id=conv_id,
+                        position=position,
+                        content_hash=content_hash,
+                        status=status,
+                        conversations=result.get('conversations', []),
+                        processing_time_ms=processing_time,
+                        tokens=tokens_total,
+                        cost=0.0,
+                        error=result.get('error')
+                    )
+                except Exception as e:
+                    logger.warning(f"Failed to record to monitor: {e}")
+            
+            # Add to checkpoint (in-memory update, actual save is batched)
+            if self.checkpoint_manager:
+                self.checkpoint_manager.add_processed(
+                    position,
+                    content_hash,
+                    status,
+                    result,
+                    save_checkpoint=False  # Batch save
+                )
+        except Exception as e:
+            logger.error(f"Error in async save: {e}", exc_info=True)
+    
+    def _async_save_checkpoint(self):
+        """
+        Async I/O: Save checkpoint without blocking LLM workers.
+        """
+        try:
+            if self.checkpoint_manager:
+                self.checkpoint_manager._save_checkpoint()
+        except Exception as e:
+            logger.error(f"Error in async checkpoint save: {e}", exc_info=True)
+    
     def _generate_parallel(self, num_conversations: int, num_workers: int):
         """Generate conversations in parallel with production features."""
         complete = 0
         partial = 0
         failed = 0
         skipped = 0
+        filtered = 0  # Track filtered conversations (quality validation failures)
         start_time = time.time()
         
         # Token tracking aggregation (for console display only)
@@ -307,53 +409,213 @@ class Runner:
             initial=complete + partial + failed + skipped
         )
         
+        # Separate executor for async I/O operations (saving, checkpointing)
+        # This ensures LLM calls never wait for disk I/O
+        io_executor = ThreadPoolExecutor(max_workers=2, thread_name_prefix="IOWorker")
+        
         with ThreadPoolExecutor(max_workers=num_workers, thread_name_prefix="Worker") as executor:
             futures = {}
+            io_futures = []  # Track I/O operations
             submitted = 0
+            conv_stream = self.data_loader.stream_conversations(skip_positions=skip_positions)
             
-            # Submit conversations using streaming loader
-            try:
-                for base_conv in self.data_loader.stream_conversations(skip_positions=skip_positions):
-                    if self.shutdown_requested:
-                        logger.warning("Shutdown requested, stopping submission")
-                        break
+            # Initial batch submission - keep executor saturated
+            initial_batch = min(num_workers * 2, num_conversations)  # 2x workers for saturation
+            for _ in range(initial_batch):
+                if self.shutdown_event.is_set():
+                    break
+                
+                try:
+                    base_conv = next(conv_stream)
+                    position = base_conv.get('_position', -1)
+                    content_hash = base_conv.get('_content_hash', '')
                     
-                    if submitted >= num_conversations:
-                        break
-                    
-                    try:
-                        position = base_conv.get('_position', -1)
-                        content_hash = base_conv.get('_content_hash', '')
-                        
-                        # Submit to thread pool
-                        future = executor.submit(
-                            self._process_conversation_with_retry,
-                            base_conv,
-                            submitted,
-                            None  # No partial resume for now
-                        )
-                        futures[future] = (submitted, position, content_hash)
-                        submitted += 1
-                        
-                    except Exception as e:
-                        logger.error(f"Error submitting conversation: {e}")
-                        
-            except Exception as e:
-                logger.error(f"Error streaming conversations: {e}")
+                    future = executor.submit(
+                        self._process_conversation_with_retry,
+                        base_conv,
+                        submitted,
+                        None  # No partial resume for now
+                    )
+                    futures[future] = (submitted, position, content_hash)
+                    submitted += 1
+                except StopIteration:
+                    break
+                except Exception as e:
+                    logger.error(f"Error submitting conversation: {e}")
             
             # Process results
             batch_count = 0
             auto_save_freq = self.config.get('checkpoint.auto_save_frequency', 10)
             
-            for future in as_completed(futures):
-                if self.shutdown_requested:
-                    logger.warning("Shutdown requested, waiting for running tasks...")
+            # Emergency shutdown: cancel all pending futures
+            if self.shutdown_event.is_set():
+                pending_count = sum(1 for f in futures if not f.done())
+                in_flight_count = len([f for f in futures if not f.done() and not f.cancelled()])
                 
+                # Update progress bar with shutdown status
+                pbar.set_description("🛑 SHUTTING DOWN")
+                pbar.set_postfix_str(f"Canceling {pending_count} tasks... ETA: ~7s")
+                
+                logger.warning(f"🛑 Emergency shutdown initiated - {pending_count} pending tasks")
+                logger.info(f"   ├─ Pending: {pending_count} | In-flight: ~{in_flight_count}")
+                logger.info(f"   └─ Estimated shutdown time: ~7 seconds")
+                
+                # Step 1: Cancel pending (< 1s)
+                pbar.set_postfix_str(f"[1/3] Canceling {pending_count} pending tasks...")
+                cancel_start = time.time()
+                for future in futures:
+                    if not future.done():
+                        future.cancel()
+                cancel_time = time.time() - cancel_start
+                logger.info(f"✓ Canceled {pending_count} tasks in {cancel_time:.2f}s")
+                
+                # Step 2: Save checkpoint (~1-2s)
+                pbar.set_postfix_str(f"[2/3] Saving checkpoint... ETA: ~2s")
+                if self.checkpoint_manager:
+                    checkpoint_start = time.time()
+                    logger.info("💾 Saving emergency checkpoint...")
+                    self.checkpoint_manager._save_checkpoint()
+                    checkpoint_time = time.time() - checkpoint_start
+                    logger.info(f"✓ Checkpoint saved in {checkpoint_time:.2f}s: {self.checkpoint_manager.checkpoint_path}")
+                
+                # Step 3: Wait for in-flight (up to 5s)
+                pbar.set_postfix_str(f"[3/3] Waiting for {in_flight_count} in-flight requests... Max 5s")
+                logger.info("⏳ Waiting up to 5 seconds for in-flight API calls...")
+                shutdown_start = time.time()
+                completed_after_shutdown = 0
+                
+                for future in as_completed(futures, timeout=5):
+                    try:
+                        elapsed = time.time() - shutdown_start
+                        remaining = max(0, 5 - elapsed)
+                        
+                        if elapsed > 5:
+                            break
+                        
+                        # Update status every 0.5s
+                        if int(elapsed * 2) != int((elapsed - 0.1) * 2):
+                            pbar.set_postfix_str(f"[3/3] Processing in-flight... {completed_after_shutdown} done, {remaining:.1f}s left")
+                        
+                        completed_after_shutdown += 1
+                    except Exception:
+                        pass
+                
+                total_shutdown_time = time.time() - cancel_start
+                logger.info(f"✓ Completed {completed_after_shutdown} in-flight requests")
+                logger.info(f"✓ Total shutdown time: {total_shutdown_time:.2f}s")
+                logger.warning("🛑 SHUTDOWN COMPLETE - Exiting now")
+                
+                pbar.set_postfix_str(f"✓ Shutdown complete in {total_shutdown_time:.2f}s")
+                pbar.close()
+                
+                # Force exit
+                sys.exit(0)
+            
+            for future in as_completed(futures):
+                if self.shutdown_event.is_set():
+                    pending_count = sum(1 for f in futures if not f.done())
+                    in_flight_count = len([f for f in futures if not f.done() and not f.cancelled()])
+                    
+                    # Update progress bar with shutdown status
+                    pbar.set_description("🛑 SHUTTING DOWN")
+                    pbar.set_postfix_str(f"Canceling {pending_count} tasks... ETA: ~7s")
+                    
+                    logger.warning(f"🛑 Emergency shutdown initiated - {pending_count} pending tasks")
+                    logger.info(f"   ├─ Pending: {pending_count} | In-flight: ~{in_flight_count}")
+                    logger.info(f"   └─ Estimated shutdown time: ~7 seconds")
+                    
+                    # Step 1: Cancel pending
+                    pbar.set_postfix_str(f"[1/3] Canceling {pending_count} pending tasks...")
+                    cancel_start = time.time()
+                    logger.warning(f"🛑 Canceling {pending_count} remaining pending tasks...")
+                    
+                    for f in futures:
+                        if not f.done():
+                            f.cancel()
+                    
+                    # Step 2: Save checkpoint
+                    pbar.set_postfix_str(f"[2/3] Saving checkpoint... ETA: ~2s")
+                    if self.checkpoint_manager:
+                        checkpoint_start = time.time()
+                        logger.info("💾 Saving emergency checkpoint...")
+                        self.checkpoint_manager._save_checkpoint()
+                        checkpoint_time = time.time() - checkpoint_start
+                        logger.info(f"✓ Checkpoint saved in {checkpoint_time:.2f}s: {self.checkpoint_manager.checkpoint_path}")
+                    
+                    # Step 3: Wait for in-flight
+                    pbar.set_postfix_str(f"[3/3] Waiting for {in_flight_count} in-flight requests... Max 5s")
+                    logger.info("⏳ Waiting up to 5 seconds for in-flight requests...")
+                    shutdown_start = time.time()
+                    completed_after_shutdown = 0
+                    
+                    for remaining_future in list(futures.keys()):
+                        if remaining_future.done():
+                            try:
+                                elapsed = time.time() - shutdown_start
+                                remaining_time = max(0, 5 - elapsed)
+                                
+                                if elapsed > 5:
+                                    break
+                                
+                                # Update status
+                                if int(elapsed * 2) != int((elapsed - 0.1) * 2):
+                                    pbar.set_postfix_str(f"[3/3] Processing in-flight... {completed_after_shutdown} done, {remaining_time:.1f}s left")
+                                
+                                # Process this completed result
+                                result = remaining_future.result()
+                                if result and not result.get('skipped'):
+                                    conv_id, position, content_hash = futures[remaining_future]
+                                    status = 'partial' if result.get('is_partial') else ('completed' if result.get('success') else 'failed')
+                                    self._async_save_result(result, status, position, content_hash)
+                                    completed_after_shutdown += 1
+                            except Exception as e:
+                                logger.debug(f"Error processing during shutdown: {e}")
+                    
+                    total_shutdown_time = time.time() - cancel_start
+                    logger.info(f"✓ Completed {completed_after_shutdown} in-flight requests")
+                    logger.info(f"✓ Total shutdown time: {total_shutdown_time:.2f}s")
+                    logger.warning("🛑 SHUTDOWN COMPLETE - Exiting now")
+                    
+                    pbar.set_postfix_str(f"✓ Shutdown complete in {total_shutdown_time:.2f}s")
+                    pbar.close()
+                    sys.exit(0)
+                
+                # IMMEDIATELY submit next item to keep workers saturated
+                # Do this BEFORE processing result to minimize idle time
+                if submitted < num_conversations and not self.shutdown_event.is_set():
+                    try:
+                        base_conv = next(conv_stream)
+                        position_next = base_conv.get('_position', -1)
+                        content_hash_next = base_conv.get('_content_hash', '')
+                        
+                        new_future = executor.submit(
+                            self._process_conversation_with_retry,
+                            base_conv,
+                            submitted,
+                            None
+                        )
+                        futures[new_future] = (submitted, position_next, content_hash_next)
+                        submitted += 1
+                    except StopIteration:
+                        pass  # No more conversations
+                    except Exception as e:
+                        logger.error(f"Error submitting next conversation: {e}")
+                
+                # Now process the completed result
                 conv_id, position, content_hash = futures[future]
                 
                 try:
                     result = future.result()
                     
+                    # Handle filtered conversations (when filter_failed_validations=True)
+                    if result is None:
+                        filtered += 1
+                        pbar.update(1)
+                        pbar.set_postfix_str(
+                            f"✓{complete} ⚠{partial} ✗{failed} ~{skipped} 🔍{filtered} | {self._get_rate_limit_metric()}"
+                        )
+                        continue
+                    
                     # Skip if marked as skipped
                     if result.get('skipped'):
                         skipped += 1
@@ -378,52 +640,33 @@ class Runner:
                         total_input_tokens += tokens.get('input_tokens', 0)
                         total_output_tokens += tokens.get('output_tokens', 0)
                     
-                    # Save to incremental saver
-                    self.incremental_saver.save_conversation(result, status=status)
+                    # ASYNC I/O: Submit saving to separate executor (non-blocking!)
+                    # This ensures LLM workers never wait for disk I/O
+                    io_future = io_executor.submit(
+                        self._async_save_result,
+                        result,
+                        status,
+                        position,
+                        content_hash,
+                        conv_id
+                    )
+                    io_futures.append(io_future)
                     
-                    # Record to monitoring
-                    if self.monitor:
-                        try:
-                            processing_time = result.get('processing_time_ms', 0)
-                            tokens_total = result.get('tokens', {}).get('total_tokens', 0)
-                            self.monitor.record_conversation(
-                                conversation_id=conv_id,
-                                position=position,
-                                content_hash=content_hash,
-                                status=status,
-                                conversations=result.get('conversations', []),
-                                processing_time_ms=processing_time,
-                                tokens=tokens_total,
-                                cost=0.0,  # Cost calculation optional, done by user
-                                error=result.get('error')
-                            )
-                        except Exception as e:
-                            logger.warning(f"Failed to record to monitor: {e}")
+                    # Batch checkpoint updates (minimal blocking)
+                    batch_count += 1
+                    if batch_count >= auto_save_freq:
+                        # Also async to avoid blocking
+                        io_executor.submit(self._async_save_checkpoint)
+                        batch_count = 0
                     
-                    # Save to checkpoint
-                    if self.checkpoint_manager:
-                        self.checkpoint_manager.add_processed(
-                            position,
-                            content_hash,
-                            status,
-                            result,
-                            save_checkpoint=False  # Batch save
-                        )
-                        batch_count += 1
-                        
-                        # Auto-save checkpoint every N conversations
-                        if batch_count >= auto_save_freq:
-                            self.checkpoint_manager._save_checkpoint()
-                            batch_count = 0
-                    
-                    # Update progress
-                    rpm = self.rate_limiter.get_rpm()
+                    # Update progress immediately (fast, non-blocking)
+                    metric_str = self._get_rate_limit_metric()
                     pbar.update(1)
                     pbar.set_postfix_str(
-                        f"✓{complete} ⚠{partial} ✗{failed} ~{skipped} | RPM:{rpm}"
+                        f"✓{complete} ⚠{partial} ✗{failed} ~{skipped} 🔍{filtered} | {metric_str}"
                     )
                     
-                    # Update monitoring
+                    # Update monitoring (lightweight)
                     if self.monitor:
                         try:
                             total_processed = complete + partial + failed + skipped
@@ -434,23 +677,23 @@ class Runner:
                 except Exception as e:
                     logger.error(f"Error processing conv {conv_id}: {e}")
                     failed += 1
-                    
-                    # Save failed to checkpoint
-                    if self.checkpoint_manager:
-                        self.checkpoint_manager.add_processed(
-                            position,
-                            content_hash,
-                            'failed',
-                            {'id': conv_id, 'error': str(e), 'conversations': []},
-                            save_checkpoint=False
-                        )
-                    
                     pbar.update(1)
             
+            # Wait for all I/O operations to complete
+            logger.info("Waiting for I/O operations to complete...")
+            for io_future in io_futures:
+                try:
+                    io_future.result(timeout=30)
+                except Exception as e:
+                    logger.error(f"I/O operation failed: {e}")
+            
             # Final checkpoint save
             if self.checkpoint_manager:
                 self.checkpoint_manager._save_checkpoint()
                 logger.info("✓ Final checkpoint saved")
+            
+            # Shutdown I/O executor
+            io_executor.shutdown(wait=True)
         
         pbar.close()
         
@@ -461,10 +704,11 @@ class Runner:
         print("\n" + "="*60)
         print("GENERATION COMPLETE")
         print("="*60)
-        print(f"✓ Complete: {complete:>5}")
-        print(f"⚠ Partial:  {partial:>5}")
-        print(f"✗ Failed:   {failed:>5}")
-        print(f"~ Skipped:  {skipped:>5}")
+        print(f"✓ Complete:  {complete:>5}")
+        print(f"⚠ Partial:   {partial:>5}")
+        print(f"✗ Failed:    {failed:>5}")
+        print(f"~ Skipped:   {skipped:>5}")
+        print(f"🔍 Filtered: {filtered:>5}  (quality validation)")
         print(f"━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━")
         print(f"💾 Saved:   {complete + partial:>5}  ({(complete+partial)/total*100:.1f}% of processed)" if total > 0 else "💾 Saved:   0")
         print(f"⏱  Time:    {total_time/60:>5.1f} min")
